import pytest

from mock import patch

from data.database import Image, RepositoryTag, ImageStorage, Repository
from data.model.repository import create_repository
from data.model.tag import (list_active_repo_tags, create_or_update_tag, delete_tag,
                            get_matching_tags, _tag_alive, get_matching_tags_for_images)
from data.model.image import find_create_or_link_image

from test.fixtures import *

def _get_expected_tags(image):
  expected_query = (RepositoryTag
                  .select()
                  .join(Image)
                  .where(RepositoryTag.hidden == False)
                  .where((Image.id == image.id) | (Image.ancestors ** ('%%/%s/%%' % image.id))))
  return set([tag.id for tag in _tag_alive(expected_query)])

@pytest.mark.parametrize('max_subqueries,max_image_lookup_count', [
  (1, 1),
  (10, 10),
  (100, 500),
])
def test_get_matching_tags(max_subqueries, max_image_lookup_count, initialized_db):
  with patch('data.model.tag._MAX_SUB_QUERIES', max_subqueries):
    with patch('data.model.tag._MAX_IMAGE_LOOKUP_COUNT', max_image_lookup_count):
      # Test for every image in the test database.
      for image in Image.select(Image, ImageStorage).join(ImageStorage):
        matching_query = get_matching_tags(image.docker_image_id, image.storage.uuid)
        matching_tags = set([tag.id for tag in matching_query])
        expected_tags = _get_expected_tags(image)
        assert matching_tags == expected_tags, "mismatch for image %s" % image.id


@pytest.mark.parametrize('max_subqueries,max_image_lookup_count', [
  (1, 1),
  (10, 10),
  (100, 500),
])
def test_get_matching_tag_ids_for_images(max_subqueries, max_image_lookup_count, initialized_db):
  with patch('data.model.tag._MAX_SUB_QUERIES', max_subqueries):
    with patch('data.model.tag._MAX_IMAGE_LOOKUP_COUNT', max_image_lookup_count):
      # Try for various sets of the first N images.
      for count in [5, 10, 15]:
        pairs = []
        expected_tags_ids = set()
        for image in Image.select(Image, ImageStorage).join(ImageStorage):
          if len(pairs) >= count:
            break

          pairs.append((image.docker_image_id, image.storage.uuid))
          expected_tags_ids.update(_get_expected_tags(image))

        matching_tags_ids = set([tag.id for tag in get_matching_tags_for_images(pairs)])
        assert matching_tags_ids == expected_tags_ids


@pytest.mark.parametrize('max_subqueries,max_image_lookup_count', [
  (1, 1),
  (10, 10),
  (100, 500),
])
def test_get_matching_tag_ids_for_all_images(max_subqueries, max_image_lookup_count, initialized_db):
  with patch('data.model.tag._MAX_SUB_QUERIES', max_subqueries):
    with patch('data.model.tag._MAX_IMAGE_LOOKUP_COUNT', max_image_lookup_count):
      pairs = []
      for image in Image.select(Image, ImageStorage).join(ImageStorage):
        pairs.append((image.docker_image_id, image.storage.uuid))

      expected_tags_ids = set([tag.id for tag in _tag_alive(RepositoryTag.select())])
      matching_tags_ids = set([tag.id for tag in get_matching_tags_for_images(pairs)])

      # Ensure every alive tag was found.
      assert matching_tags_ids == expected_tags_ids


def test_get_matching_tag_ids_images_filtered(initialized_db):
  def filter_query(query):
    return query.join(Repository).where(Repository.name == 'simple')

  filtered_images = filter_query(Image
                                 .select(Image, ImageStorage)
                                 .join(RepositoryTag)
                                 .switch(Image)
                                 .join(ImageStorage)
                                 .switch(Image))

  expected_tags_query = _tag_alive(filter_query(RepositoryTag
                                                .select()))

  pairs = []
  for image in filtered_images:
    pairs.append((image.docker_image_id, image.storage.uuid))

  matching_tags = get_matching_tags_for_images(pairs, filter_images=filter_query,
                                               filter_tags=filter_query)

  expected_tag_ids = set([tag.id for tag in expected_tags_query])
  matching_tags_ids = set([tag.id for tag in matching_tags])

  # Ensure every alive tag was found.
  assert matching_tags_ids == expected_tag_ids


def assert_tags(repository, *args):
  tags = list(list_active_repo_tags(repository))
  assert len(tags) == len(args)

  tags_dict = {}
  for tag in tags:
    assert not tag.name in tags_dict
    assert not tag.hidden
    assert not tag.lifetime_end_ts

    tags_dict[tag.name] = tag

  for expected in args:
    assert expected in tags_dict

def test_list_active_tags(initialized_db):
  # Create a new repository.
  repository = create_repository('devtable', 'somenewrepo', None)

  # Create some images.
  image1 = find_create_or_link_image('foobarimage1', repository, None, {}, 'local_us')
  image2 = find_create_or_link_image('foobarimage2', repository, None, {}, 'local_us')

  # Make sure its tags list is empty.
  assert_tags(repository)

  # Add some new tags.
  footag = create_or_update_tag('devtable', 'somenewrepo', 'foo', image1.docker_image_id)
  bartag = create_or_update_tag('devtable', 'somenewrepo', 'bar', image1.docker_image_id)

  # Since timestamps are stored on a second-granuality, we need to make the tags "start"
  # before "now", so when we recreate them below, they don't conflict.
  footag.lifetime_start_ts -= 5
  bartag.lifetime_start_ts -= 5

  footag.save()
  bartag.save()

  # Make sure they are returned.
  assert_tags(repository, 'foo', 'bar')

  # Delete a tag and make sure it isn't returned.
  footag = delete_tag('devtable', 'somenewrepo', 'foo')
  footag.lifetime_end_ts -= 4
  footag.save()

  assert_tags(repository, 'bar')

  # Add a new foo again.
  footag = create_or_update_tag('devtable', 'somenewrepo', 'foo', image1.docker_image_id)
  footag.lifetime_start_ts -= 3
  footag.save()

  assert_tags(repository, 'foo', 'bar')

  # "Move" foo by updating it and make sure we don't get duplicates.
  create_or_update_tag('devtable', 'somenewrepo', 'foo', image2.docker_image_id)
  assert_tags(repository, 'foo', 'bar')
